/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is cmgui_corba_bridge.
 *
 * The Initial Developer of the Original Code is
 * Auckland Uniservices Ltd, Auckland, New Zealand.
 * Portions created by the Initial Developer are Copyright (C) 2005
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
#include "ComputedFieldImpl.hpp"
#include "NodeImpl.hpp"
#include "NodeFieldCreatorImpl.hpp"
#include "TimeSequenceImpl.hpp"
extern "C"
{
#include <computed_field/computed_field.h>
#include <computed_field/computed_field_finite_element.h>
#include <api/cmiss_time_sequence.h>
#include <api/cmiss_core.h>
}

extern int server_closing;
extern CORBA::ORB_var orb;

  CMGUIAPI::ComputedField::DoubleSeq*
  ComputedFieldImpl::evaluateAtNode(
	  const CMGUIAPI::Node_ptr nodewrap,
	  const CORBA::Double time)
{
	int return_code;

	PortableServer::ServantBase* sbnode = poa->reference_to_servant(nodewrap);
	Cmiss_node_id node = dynamic_cast<NodeImpl*>(sbnode)->impl;

	unsigned int i;
	unsigned int number_of_components = Computed_field_get_number_of_components(impl);
	FE_value values[number_of_components];

	return_code = Computed_field_evaluate_at_node(impl, node,
		(FE_value)time, values);
		
	if (!return_code)
      throw CMGUIAPI::CmguiError();
   
	CMGUIAPI::ComputedField::DoubleSeq_var seq =
		new CMGUIAPI::ComputedField::DoubleSeq();

	seq->length(number_of_components);
	for (i = 0 ; i < number_of_components ; i++)
	{
		seq[i] = values[i];
	}
	return(seq._retn());
}

  char*
  ComputedFieldImpl::evaluateAsStringAtNode(
	  const CMGUIAPI::Node_ptr nodewrap,
	  const CORBA::Double time)
{
	char *cmgui_string, *return_string;

	PortableServer::ServantBase* sbnode = poa->reference_to_servant(nodewrap);
	Cmiss_node_id node = dynamic_cast<NodeImpl*>(sbnode)->impl;

	cmgui_string = Computed_field_evaluate_as_string_at_node(
		impl, /*All components*/-1, node, (FE_value)time);
		
	if (!cmgui_string)
      throw CMGUIAPI::CmguiError();
   
	return_string = CORBA::string_dup(cmgui_string);
	Cmiss_deallocate(cmgui_string);

	return(return_string);
}

  void
  ComputedFieldImpl::setValuesAtNode(
	  const CMGUIAPI::Node_ptr nodewrap,
	  const CORBA::Double time,
	  const CMGUIAPI::ComputedField::DoubleSeq& values_sequence)
{
	int return_code;

	PortableServer::ServantBase* sbnode = poa->reference_to_servant(nodewrap);
	Cmiss_node_id node = dynamic_cast<NodeImpl*>(sbnode)->impl;

	unsigned int i;
	unsigned int number_of_components = Computed_field_get_number_of_components(impl);

	if (number_of_components != values_sequence.length())
      throw CMGUIAPI::CmguiError();

	FE_value values[number_of_components];
	for (i = 0 ; i < number_of_components ; i++)
	{
		values[i] = values_sequence[i];
	}

	return_code = Computed_field_set_values_at_node(impl, node,
		(FE_value)time, values);
		
	if (!return_code)
      throw CMGUIAPI::CmguiError();
   
}

  CORBA::Boolean
  ComputedFieldImpl::isDefinedAtNode(
	  const CMGUIAPI::Node_ptr nodewrap)
{
	int return_code;

	PortableServer::ServantBase* sbnode = poa->reference_to_servant(nodewrap);
	Cmiss_node_id node = dynamic_cast<NodeImpl*>(sbnode)->impl;

	return_code = Computed_field_is_defined_at_node(impl, node);
		
	if (return_code)
		return true;
	else
		return false;
   
}

  void
  ComputedFieldImpl::defineAtNode(
	  const CMGUIAPI::Node_ptr nodewrap,
	  const CMGUIAPI::TimeSequence_ptr timesequencewrap,
	  const CMGUIAPI::NodeFieldCreator_ptr nodefieldcreatorwrap)
{
	Cmiss_time_sequence_id ctimeseq = NULL;
	Cmiss_node_field_creator_id cnfc = NULL;
	PortableServer::ServantBase* sbtimeseq = NULL;
	PortableServer::ServantBase_var sbtimeseqv;
	PortableServer::ServantBase* sbnfc = NULL;
	PortableServer::ServantBase_var sbnfcv;
	int return_code;

	PortableServer::ServantBase* sbnode = poa->reference_to_servant(nodewrap);
	Cmiss_node_id node = dynamic_cast<NodeImpl*>(sbnode)->impl;

	if (!Computed_field_is_type_finite_element(impl))
      throw CMGUIAPI::CmguiError();

	if (!CORBA::is_nil(timesequencewrap))
	{
		sbtimeseq = poa->reference_to_servant(timesequencewrap);
		sbtimeseqv = sbtimeseq;
		ctimeseq = dynamic_cast<TimeSequenceImpl*>(sbtimeseq)->impl;
	}
	
	if (!CORBA::is_nil(nodefieldcreatorwrap))
	{
		sbnfc = poa->reference_to_servant(nodefieldcreatorwrap);
		sbnfcv = sbnfc;
		cnfc = dynamic_cast<NodeFieldCreatorImpl*>(sbnfc)->impl;
	}

	return_code = Computed_field_finite_element_define_at_node(
		impl, node, ctimeseq, cnfc);
		
	if (!return_code)
      throw CMGUIAPI::CmguiError();
   
}

  void
  ComputedFieldImpl::setStringAtNode(
	  const CORBA::Long component_number,
	  const CMGUIAPI::Node_ptr nodewrap,
	  const CORBA::Double time,
	  const char *string_value)
{
	int return_code;

	PortableServer::ServantBase* sbnode = poa->reference_to_servant(nodewrap);
	Cmiss_node_id node = dynamic_cast<NodeImpl*>(sbnode)->impl;

	if (!Computed_field_is_type_finite_element(impl))
      throw CMGUIAPI::CmguiError();

	return_code = Computed_field_finite_element_set_string_at_node(
		impl, component_number, node, time, const_cast<char*>(string_value));
		
	if (!return_code)
      throw CMGUIAPI::CmguiError();
   
}
