/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is cmgui_corba_bridge.
 *
 * The Initial Developer of the Original Code is
 * Auckland Uniservices Ltd, Auckland, New Zealand.
 * Portions created by the Initial Developer are Copyright (C) 2005
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
#include "FunctionImpl.hpp"
extern "C"
{
#include <api/cmiss_finite_element.h>
#include <api/cmiss_function_finite_element.h>
#include <api/cmiss_time_sequence.h>
#include <api/cmiss_function_variable.h>
}
#include "NodeImpl.hpp"
#include "TimeSequenceImpl.hpp"
#include "NodeFieldCreatorImpl.hpp"
#include "ElementImpl.hpp"
#include "FunctionVariableImpl.hpp"
#include <stdio.h>

void
FunctionImpl::defineOnNode
(
 CMGUIAPI::Node_ptr node,
 CMGUIAPI::TimeSequence_ptr timeSeq,
 CMGUIAPI::NodeFieldCreator_ptr creator
)
{
  /* If we get an exception like WrongAdaptor on fetching servant, just pass
   * this back to the caller.
   */
  Cmiss_node_id cnode = NULL;
  Cmiss_time_sequence_id ctimeseq = NULL;
  Cmiss_node_field_creator_id cnfc = NULL;
  PortableServer::ServantBase* sbnode = NULL;
  PortableServer::ServantBase_var sbnodev;
  PortableServer::ServantBase* sbtimeseq = NULL;
  PortableServer::ServantBase_var sbtimeseqv;
  PortableServer::ServantBase* sbnfc = NULL;
  PortableServer::ServantBase_var sbnfcv;

  if (!CORBA::is_nil(node))
  {
    sbnode = poa->reference_to_servant(node);
    sbnodev = sbnode;
    cnode = dynamic_cast<NodeImpl*>(sbnode)->impl;
  }

  if (!CORBA::is_nil(timeSeq))
  {
    sbtimeseq = poa->reference_to_servant(timeSeq);
    sbtimeseqv = sbtimeseq;
    ctimeseq = dynamic_cast<TimeSequenceImpl*>(sbtimeseq)->impl;
  }

  if (!CORBA::is_nil(creator))
  {
    sbnfc = poa->reference_to_servant(creator);
    sbnfcv = sbnfc;
    cnfc = dynamic_cast<NodeFieldCreatorImpl*>(sbnfc)->impl;
  }

  Cmiss_function_finite_element_define_on_Cmiss_node
  (
   impl,
   cnode,
   ctimeseq,
   cnfc
  );
}

void
FunctionImpl::defineTensorProductBasisOnElement
(
 CMGUIAPI::Element_ptr el,
 CORBA::Long dimension,
 CORBA::ULong basis_type
)
{
  enum Cmiss_basis_type basis;
  switch (basis_type)
  {
  case CMGUIAPI::Function::BASIS_NO_RELATION:
    basis = NO_RELATION;
    break;
  case CMGUIAPI::Function::BASIS_BSPLINE:
    basis = BSPLINE;
    break;
  case CMGUIAPI::Function::BASIS_CUBIC_HERMITE:
    basis = CUBIC_HERMITE;
    break;
  case CMGUIAPI::Function::BASIS_CUBIC_LAGRANGE:
    basis = CUBIC_LAGRANGE;
    break;
  case CMGUIAPI::Function::BASIS_FOURIER:
    basis = FOURIER;
    break;
  case CMGUIAPI::Function::BASIS_HERMITE_LAGRANGE:
    basis = HERMITE_LAGRANGE;
    break;
  case CMGUIAPI::Function::BASIS_LAGRANGE_HERMITE:
    basis = LAGRANGE_HERMITE;
    break;
  case CMGUIAPI::Function::BASIS_LINEAR_LAGRANGE:
    basis = LINEAR_LAGRANGE;
    break;
  case CMGUIAPI::Function::BASIS_LINEAR_SIMPLEX:
    basis = LINEAR_SIMPLEX;
    break;
  case CMGUIAPI::Function::BASIS_POLYGON:
    basis = POLYGON;
    break;
  case CMGUIAPI::Function::BASIS_QUADRATIC_LAGRANGE:
    basis = QUADRATIC_LAGRANGE;
    break;
  case CMGUIAPI::Function::BASIS_QUADRATIC_SIMPLEX:
    basis = QUADRATIC_SIMPLEX;
    break;
  case CMGUIAPI::Function::BASIS_SERENDIPITY:
    basis = SERENDIPITY;
    break;
  case CMGUIAPI::Function::BASIS_SINGULAR:
    basis = SINGULAR;
    break;
  case CMGUIAPI::Function::BASIS_TRANSITION:
    basis = TRANSITION;
    break;
  default:
    throw CMGUIAPI::CmguiError();
  }

  ServantBase* sbel = poa->reference_to_servant(el);
  Cmiss_element_id element = dynamic_cast<ElementImpl*>(sbel)->impl;

  if (!Cmiss_function_finite_element_define_tensor_product_basis_on_element
      (
       impl,
       element,
       dimension, basis
      )
     )
  {
    sbel->_remove_ref();
    throw CMGUIAPI::CmguiError();
  }

  sbel->_remove_ref();
}

CMGUIAPI::FunctionVariable_ptr
FunctionImpl::getNodalValues
(
 const char* componentName,
 CORBA::Long componentIdentifier,
 CMGUIAPI::Node_ptr n,
 CORBA::ULong valuesType,
 CORBA::Long version,
 CMGUIAPI::TimeSequence_ptr timeSeq
)
{
  enum Cmiss_nodal_value_type valueType;

  switch (valuesType)
  {
  case CMGUIAPI::Function::NODAL_VALUE:
    valueType = FE_NODAL_VALUE;
    break;
  case CMGUIAPI::Function::NODAL_D_DS1:
    valueType = FE_NODAL_D_DS1;
    break;
  case CMGUIAPI::Function::NODAL_D_DS2:
    valueType = FE_NODAL_D_DS2;
    break;
  case CMGUIAPI::Function::NODAL_D_DS3:
    valueType = FE_NODAL_D_DS3;
    break;
  case CMGUIAPI::Function::NODAL_D2_DS1DS2:
    valueType = FE_NODAL_D2_DS1DS2;
    break;
  case CMGUIAPI::Function::NODAL_D2_DS1DS3:
    valueType = FE_NODAL_D2_DS1DS3;
    break;
  case CMGUIAPI::Function::NODAL_D2_DS2DS3:
    valueType = FE_NODAL_D2_DS2DS3;
    break;
  case CMGUIAPI::Function::NODAL_D3_DS1DS2DS3:
    valueType = FE_NODAL_D3_DS1DS2DS3;
    break;
  default:
    printf("Warning: Invalid valueType!\n");
    throw CMGUIAPI::CmguiError();
  }

  PortableServer::ServantBase* sbnode = poa->reference_to_servant(n);
  PortableServer::ServantBase_var sbnodev(sbnode);
  Cmiss_node_id node = dynamic_cast<NodeImpl*>(sbnode)->impl;

  Cmiss_time_sequence_id timeseq = NULL;
  PortableServer::ServantBase* sbtimeseq = NULL;
  PortableServer::ServantBase_var sbtimeseqv;
  if (!CORBA::is_nil(timeSeq))
  {
    sbtimeseq = poa->reference_to_servant(timeSeq);
    sbtimeseqv = sbtimeseq;
    timeseq = dynamic_cast<TimeSequenceImpl*>(sbtimeseq)->impl;
  }

  Cmiss_function_variable_id fv =
    Cmiss_function_finite_element_nodal_values
    (
     impl, const_cast<char*>(componentName),
     componentIdentifier,
     node, valueType, version, timeseq
    );

  if (fv == NULL)
    throw CMGUIAPI::CmguiError();

  FunctionVariableImpl* fvi = new FunctionVariableImpl(fv, poa);
  delete poa->activate_object(fvi);
  CMGUIAPI::FunctionVariable_ptr fvp = fvi->_this();
  fvi->_remove_ref();
  return fvp;
}
