/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is cmgui_corba_bridge.
 *
 * The Initial Developer of the Original Code is
 * Auckland Uniservices Ltd, Auckland, New Zealand.
 * Portions created by the Initial Developer are Copyright (C) 2005
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
#include <sys/types.h>
extern "C"
{
#include <api/cmiss_scene_viewer.h>
}
#include "interfaces/SceneViewer.hh"
#ifdef WIN32
#include <windows.h>
#else
#include <gtk/gtk.h>
#endif
#include "Utilities.hpp"

#include <stdio.h>

class SceneViewerImpl
  : public POA_CMGUIAPI::SceneViewer,
    public virtual PortableServer::RefCountServantBase
{
public:
  IMPL_XPCOM1(CMGUIAPI::SceneViewer);

#ifdef WIN32
  SceneViewerImpl(
                  HWND ichildHWND,
                  Cmiss_scene_viewer_id iimpl,
                  PortableServer::POA_ptr ipoa
                 )
    :  refcount(1),impl(iimpl), poa(PortableServer::POA::_duplicate(ipoa)),
      childHWND(ichildHWND) {}
#else
  SceneViewerImpl(GtkWidget* ipackbox,
                  Cmiss_scene_viewer_id iimpl,
                  PortableServer::POA_ptr ipoa)
    : refcount(1), impl(iimpl), poa(PortableServer::POA::_duplicate(ipoa)), packbox(ipackbox) {}
#endif

  virtual ~SceneViewerImpl()
  {
#ifdef WIN32
    DESTROY(Cmiss_scene_viewer)(&impl);
    DestroyWindow(childHWND);
#else
    /* We don't need to do this, it should already be done.
     * gtk_widget_destroy(packbox); */
#endif
  }


  void close();

  void getLookatParameters
  (
   CORBA::Double& eyeX,
   CORBA::Double& eyeY,
   CORBA::Double& eyeZ,
   CORBA::Double& lookAtX,
   CORBA::Double& lookAtY,
   CORBA::Double& lookAtZ,
   CORBA::Double& upX,
   CORBA::Double& upY,
   CORBA::Double& upZ
  );
  void setLookatParametersNonSkew
  (
   CORBA::Double eyeX,
   CORBA::Double eyeY,
   CORBA::Double eyeZ,
   CORBA::Double lookAtX,
   CORBA::Double lookAtY,
   CORBA::Double lookAtZ,
   CORBA::Double upX,
   CORBA::Double upY,
   CORBA::Double upZ
  );
  void getNearAndFarPlane
  (
   CORBA::Double& nearPlane,
   CORBA::Double& farPlane
  );
  void setNearAndFarPlane
  (
   CORBA::Double nearPlane,
   CORBA::Double farPlane
  );
  void getBackgroundColourRGB
  (
   CORBA::Double& red,
   CORBA::Double& green,
   CORBA::Double& blue
  );
  void setBackgroundColourRGB
  (
   CORBA::Double red,
   CORBA::Double green,
   CORBA::Double blue
  );
  void setSceneByName(const char* sceneName);
  void setOverlaySceneByName(const char* sceneName);
  void setBackgroundTextureByName(const char* textureName);
  void setBackgroundTexturePosition
  (
   CORBA::Double left,
   CORBA::Double right,
   CORBA::Double width,
   CORBA::Double height
  );
  void setInteractiveToolByName(const char* toolName);
  void viewAll();
  void redrawNow();
  void getFreespinTumbleAxis(CORBA::Float& axisX, CORBA::Float& axisY, CORBA::Float& axisZ);
  void startFreespin(CORBA::Float tumbleAxisX, CORBA::Float tumbleAxisY,
                     CORBA::Float tumbleAxisZ, CORBA::Double tumbleAngle);
  void stopAnimations();
  void writeImageToFile(const char* fileName, CORBA::Boolean forceOnscreen,
                        CORBA::ULong preferredWidth,
                        CORBA::ULong preferredHeight,
                        CORBA::ULong preferredAntiAlias,
                        CORBA::ULong preferredTransparencyLayers);
  void openSceneEditor();
  void setWindowSize(CORBA::ULong width, CORBA::ULong height);
  CORBA::ULong projectionMode();
  void projectionMode(CORBA::ULong projMode);
  CORBA::ULong viewportMode();
  void viewportMode(CORBA::ULong viewportMode);
  CORBA::ULong transparencyMode();
  void transparencyMode(CORBA::ULong transMode);
  CORBA::ULong transparencyLayers();
  void transparencyLayers(CORBA::ULong transLayers);
  CORBA::Double viewAngle();
  void viewAngle(CORBA::Double angle);
  CORBA::ULong antialiasMode();
  void antialiasMode(CORBA::ULong aaMode);
  void getDepthOfField
  (
   CORBA::Double& depthOfField,
   CORBA::Double& focalDepth
  );
  void setDepthOfField
  (
   CORBA::Double depthOfField,
   CORBA::Double focalDepth
  );
  CORBA::Boolean perturbLines();
  void perturbLines(CORBA::Boolean pLines);
  CORBA::Double translationRate();
  void translationRate(CORBA::Double rate);
  CORBA::Double tumbleRate();
  void tumbleRate(CORBA::Double rate);
  CORBA::Double zoomRate();
  void zoomRate(CORBA::Double rate);
  char* sceneName();
  CORBA::Double freespinTumbleAngle();
  void freespinTumbleAngle(CORBA::Double rate);
  void getViewingVolume(CORBA::Double& left, CORBA::Double& right,
                        CORBA::Double& bottom, CORBA::Double& top,
                        CORBA::Double& near, CORBA::Double& far);
  void setViewingVolume(CORBA::Double left, CORBA::Double right,
                        CORBA::Double bottom, CORBA::Double top,
                        CORBA::Double near, CORBA::Double far);
  Cmiss_scene_viewer_id impl;
private:
  PortableServer::POA_var poa;
#ifdef WIN32
  HWND childHWND;
#else
  GtkWidget* packbox;
#endif
};
