/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is cmgui_corba_bridge.
 *
 * The Initial Developer of the Original Code is
 * Auckland Uniservices Ltd, Auckland, New Zealand.
 * Portions created by the Initial Developer are Copyright (C) 2005
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
#include "CommandDataFactoryImpl.hpp"
#include <iostream>
extern "C" {
#include <command/cmiss.h>
};
#include <api/cmiss_timer.h>
#include <omniORB4/poa.h>
#include "ApplicationBootstrap.hh"
#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#ifndef WIN32
#include <sys/time.h>
#endif
#include <inttypes.h>

static void CreatePermanentObjects(CORBA::ORB_ptr orb, PortableServer::POA_ptr poa,
                                   CMGUIAPI::ApplicationBootstrap_ptr bootstrap);

int server_closing = 0;

struct Cmiss_command_data* the_command_data;
CORBA::ORB_var orb;

static unsigned long sleep_time = 500000000;

struct exception_handler_chain
{
  struct exception_handler_chain* parent;
  void* handler_function;
};

int
do_corba_work(void *user_data)
{
  //printf("do_corba_work entered.\n");
#ifdef WIN32
  // We need to turn off Win32 exception handling...
  struct exception_handler_chain *eh_top_original, *eh_bottom, *eh_tmp;
  asm("fs movl (0), %0" : "=r"(eh_top_original));
  eh_tmp = eh_bottom = eh_top_original;
  while (eh_tmp != NULL &&
	 eh_tmp != (void*)0xFFFFFFFF)
  {
    eh_bottom = eh_tmp;
    eh_tmp = eh_tmp->parent;
  }
  asm("fs movl %0, (0)" : : "r"(eh_bottom));
#endif
  bool did_work = false;
  Cmiss_timer_package_id idlepkg = (Cmiss_timer_package_id)user_data;
  while (orb->work_pending())
  {
    did_work = true;
    orb->perform_work();
  }
  if (server_closing)
  {
    orb->shutdown(true);
#ifdef WIN32
    asm("fs movl %0, (0)" : : "r"(eh_top_original));
#endif
    exit(0);
  }
  //printf("Adding timer package callback for %lu ns.\n", sleep_time);
  Cmiss_timer_package_add_callback(idlepkg, 0, sleep_time, do_corba_work, idlepkg);
  /* XXX but how else do we do this? */
  if (!did_work)
  {
    sleep_time *= 2;
    if (sleep_time > 20000000)
      sleep_time = 20000000;
  }
  else
  {
    sleep_time /= 2;
    if (sleep_time < 1000000)
      sleep_time = 1000000;
  }
#ifdef WIN32
  asm("fs movl %0, (0)" : : "r"(eh_top_original));
#endif
  return 0;
}

int
main(int argc, char** argv)
{
  if (argc < 2)
  {
    printf("Usage: corba_bridge application_bootstrap_ior\n");
    return -1;
  }

#ifdef WIN32
  /* Need IONBF, since \n is used as a line terminator but MSVCRT doesn't
   * flush the buffer until it gets \r\n.
   */
  setvbuf(stdout, NULL, _IONBF, 0);
  setvbuf(stderr, NULL, _IONBF, 0);
#endif

  try
  {
    /* Initialise the ORB... */
    char* orb_argv[] =
      {
	"cmgui_corba_bridge",
#ifdef WIN32
	"-ORBendPoint", "giop:tcp:127.0.0.1:0",
#else
	"-ORBendPoint", "giop:unix:",
#endif
	NULL
      };
    int orb_argc = sizeof(orb_argv) / sizeof(char*) - 1;
    orb = CORBA::ORB_init(orb_argc, orb_argv);

    CORBA::Object_var appbootstrapObj = orb->string_to_object(argv[1]);
    if (CORBA::is_nil(appbootstrapObj))
    {
      printf("You passed an invalid IOR on the command-line.\n");
      return -1;
    }

    CMGUIAPI::ApplicationBootstrap_var appbootstrap =
      CMGUIAPI::ApplicationBootstrap::_narrow(appbootstrapObj);
    if (CORBA::is_nil(appbootstrap))
    {
      printf("The appbootstrap must implement CMGUIAPI::ApplicationBootstrap(%s doesn't)!\n", argv[1]);
      return -1;
    }

    /* Fetch the root POA... */
    CORBA::Object_var obj = orb->resolve_initial_references("RootPOA");
    PortableServer::POA_var poa = PortableServer::POA::_narrow(obj);
    
    /* Now make a special MAIN_THREAD POA... */
    CORBA::PolicyList policylist;
    policylist.length(1);
    policylist[0] = poa->create_thread_policy
                    (
                     PortableServer::MAIN_THREAD_MODEL
                    );
    poa = poa->create_POA("MainThreadPOA", poa->the_POAManager(),
                          policylist);

    /* Activate the POA... */
    PortableServer::POAManager_var pman = poa->the_POAManager();
    pman->activate();

    /* Set up all factories... This is also where we notify the cmgui is ready,
     * so from here on, calls will start to happen...
     */
    CreatePermanentObjects(orb, poa, appbootstrap);

    Cmiss_timer_package_id idlepkg
      = Cmiss_command_data_get_timer_package(the_command_data);
    //printf("Adding first timer callback for %lu nanoseconds.\n", sleep_time);
    Cmiss_timer_package_add_callback(idlepkg, 0, sleep_time, do_corba_work, idlepkg);

    /* Go to the cmgui main-loop...
     */
    Cmiss_command_data_main_loop(the_command_data);
  }
  catch (const CORBA::SystemException& e)
  {
#undef minor
    std::cerr << "A CORBA exception occurred while starting the server: "
              << e._name()
              << " minor " << (e.minor())
              << std::endl;
  }
  catch (...)
  {
    std::cerr << "An unsupported exception occurred while starting the server."
              << std::endl;
  }
  return 0;
}

void
CreatePermanentObjects(CORBA::ORB_ptr, PortableServer::POA_ptr poa,
                       CMGUIAPI::ApplicationBootstrap_ptr bootstrap)
{
  CommandDataFactoryImpl* cdfi = new CommandDataFactoryImpl(poa);
  the_command_data = cdfi->GetCMGUICommandData();
  PortableServer::ObjectId_var oid = poa->activate_object(cdfi);

  bootstrap->cmguiReady(cdfi->_this());
  bootstrap->release_ref();
  cdfi->_remove_ref();

#ifdef WIN32
  WNDCLASS srvWC;

  srvWC.style = CS_HREDRAW | CS_VREDRAW;
  srvWC.lpfnWndProc = DefWindowProc;
  srvWC.cbClsExtra = 0;
  srvWC.cbWndExtra = 0;
  srvWC.hInstance = GetModuleHandle(NULL);
  srvWC.hIcon = NULL;
  srvWC.hCursor = LoadCursor(NULL, IDC_ARROW);
  srvWC.hbrBackground = (HBRUSH)COLOR_BACKGROUND;
  srvWC.lpszMenuName = NULL;
  srvWC.lpszClassName = "cmgui_embed_wndclass";
  RegisterClass(&srvWC);
#endif
}

/* 
   The following license applies to the original Mersenne Twister code. My
   modifications are Copyright (C) 2006, and are under the same license as the
   rest of the code.

   A C-program for MT19937, with initialization improved 2002/1/26.
   Coded by Takuji Nishimura and Makoto Matsumoto.

   Before using, initialize the state by using init_genrand(seed)  
   or init_by_array(init_key, key_length).

   Copyright (C) 1997 - 2002, Makoto Matsumoto and Takuji Nishimura,
   All rights reserved.                          

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:

     1. Redistributions of source code must retain the above copyright
        notice, this list of conditions and the following disclaimer.

     2. Redistributions in binary form must reproduce the above copyright
        notice, this list of conditions and the following disclaimer in the
        documentation and/or other materials provided with the distribution.

     3. The names of its contributors may not be used to endorse or promote 
        products derived from this software without specific prior written 
        permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


   Any feedback is very welcome.
   http://www.math.sci.hiroshima-u.ac.jp/~m-mat/MT/emt.html
   email: m-mat @ math.sci.hiroshima-u.ac.jp (remove space)
*/

#define N 624
#define M 397
#define MATRIX_A 0x9908b0dfUL   /* constant vector a */
#define UPPER_MASK 0x80000000UL /* most significant w-r bits */
#define LOWER_MASK 0x7fffffffUL /* least significant r bits */

static unsigned long mt[N];
static int mti=N+1;

/* initializes mt[N] with a seed */
static void
mersenne_init_genrand(unsigned long s)
{
    mt[0]= s & 0xffffffffUL;
    for (mti=1; mti<N; mti++) {
        mt[mti] = 
	    (1812433253UL * (mt[mti-1] ^ (mt[mti-1] >> 30)) + mti); 
        /* See Knuth TAOCP Vol2. 3rd Ed. P.106 for multiplier. */
        /* In the previous versions, MSBs of the seed affect   */
        /* only MSBs of the array mt[].                        */
        /* 2002/01/09 modified by Makoto Matsumoto             */
        mt[mti] &= 0xffffffffUL;
        /* for >32 bit machines */
    }
}

static void
mersenne_init_by_array(unsigned long init_key[], int key_length)
{
    int i, j, k;
    mersenne_init_genrand(19650218UL);
    i=1; j=0;
    k = (N>key_length ? N : key_length);
    for (; k; k--) {
        mt[i] = (mt[i] ^ ((mt[i-1] ^ (mt[i-1] >> 30)) * 1664525UL))
          + init_key[j] + j; /* non linear */
        mt[i] &= 0xffffffffUL; /* for WORDSIZE > 32 machines */
        i++; j++;
        if (i>=N) { mt[0] = mt[N-1]; i=1; }
        if (j>=key_length) j=0;
    }
    for (k=N-1; k; k--) {
        mt[i] = (mt[i] ^ ((mt[i-1] ^ (mt[i-1] >> 30)) * 1566083941UL))
          - i; /* non linear */
        mt[i] &= 0xffffffffUL; /* for WORDSIZE > 32 machines */
        i++;
        if (i>=N) { mt[0] = mt[N-1]; i=1; }
    }

    mt[0] = 0x80000000UL; /* MSB is 1; assuring non-zero initial array */ 
}

static void
mersenne_autoseed(void)
{
  unsigned long key[50];
  unsigned long *p = key;
#ifndef WIN32
  *p++ = getuid();
  *p++ = getpid();
  struct timeval tv;
  gettimeofday(&tv, NULL);
  *p++ = tv.tv_sec;
  *p++ = tv.tv_usec;
  memset(key, 0, 50 * sizeof(unsigned long));
  gethostname((char*)p, (p - key) * 4);
  key[49] = 0;
  uint32_t l = strlen((char*)p);
  p += (l>>2) + ((l & 3) ? 1 : 0);
#else
  GetSystemTimeAsFileTime((LPFILETIME)p);
  p += 2;
  *p++ = GetCurrentProcessId();
#endif  
  mersenne_init_by_array(key, p - key);
}

/* generates a random number on [0,0xffffffff]-interval */
unsigned long mersenne_genrand_int32(void)
{
    unsigned long y;
    static unsigned long mag01[2]={0x0UL, MATRIX_A};
    /* mag01[x] = x * MATRIX_A  for x=0,1 */

    if (mti >= N) { /* generate N words at one time */
        int kk;

        if (mti == N+1)
          /* if init_genrand() has not been called, a default initial seed is
           * used. Note that this seed has changed in the Bioengineering
           * Institute version to include the time, pid, and hostname.
           */
          mersenne_autoseed();

        for (kk=0;kk<N-M;kk++) {
            y = (mt[kk] & UPPER_MASK) | (mt[kk+1] & LOWER_MASK);
            mt[kk] = mt[kk+M] ^ (y >> 1) ^ mag01[y & 0x1UL];
        }
        for (;kk<N-1;kk++) {
            y = (mt[kk] & UPPER_MASK) | (mt[kk+1] & LOWER_MASK);
            mt[kk] = mt[kk+(M-N)] ^ (y >> 1) ^ mag01[y & 0x1UL];
        }
        y = (mt[N-1]&UPPER_MASK)|(mt[0]&LOWER_MASK);
        mt[N-1] = mt[M-1] ^ (y >> 1) ^ mag01[y & 0x1UL];

        mti = 0;
    }
  
    y = mt[mti++];

    /* Tempering */
    y ^= (y >> 11);
    y ^= (y << 7) & 0x9d2c5680UL;
    y ^= (y << 15) & 0xefc60000UL;
    y ^= (y >> 18);

    return y;
}
